/*
 * Copyright 2023 Red Hat, Inc.
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: MIT
 */

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <linux/dma-buf.h>
#include <sys/mman.h>
#include <sys/types.h>

#include <xf86drm.h>

#include "virgl_context.h"
#include "virgl_util.h"
#include "virglrenderer.h"

#include "util/anon_file.h"
#include "util/hash_table.h"
#include "util/u_atomic.h"

#include "drm_context.h"
#include "drm_fence.h"
#include "drm_object.h"

#include "asahi_drm.h"
#include "asahi_proto.h"
#include "asahi_renderer.h"

static unsigned nr_timelines;

static int
gem_close(int fd, uint32_t handle)
{
   struct drm_gem_close args = { .handle = handle };
   return drmIoctl(fd, DRM_IOCTL_GEM_CLOSE, &args);
}

/**
 * Probe capset params.
 */
int
asahi_renderer_probe(UNUSED int fd, struct virgl_renderer_capset_drm *capset)
{
   drm_log("");

   capset->wire_format_version = 2;

   nr_timelines = 1;

   return 0;
}

static void
resource_delete_fxn(struct hash_entry *entry)
{
   free((void *)entry->data);
}

static void
asahi_renderer_destroy(struct virgl_context *vctx)
{
   struct drm_context *dctx = to_drm_context(vctx);

   for (unsigned i = 0; i < nr_timelines; i++)
      drm_timeline_fini(&dctx->timelines[i]);

   close(dctx->eventfd);

   drm_renderer_unmap_blob(dctx);

   _mesa_hash_table_destroy(dctx->resource_table, resource_delete_fxn);
   _mesa_hash_table_destroy(dctx->blob_table, resource_delete_fxn);
   _mesa_hash_table_destroy(dctx->sq_to_ring_idx_table, NULL);

   close(dctx->fd);
   free(dctx);
}

static void
asahi_renderer_detach_resource(struct virgl_context *vctx, struct virgl_resource *res)
{
   drm_renderer_detach_resource(vctx, res, gem_close);
}

static int
asahi_renderer_transfer_3d(UNUSED struct virgl_context *vctx,
                           UNUSED struct virgl_resource *res,
                           UNUSED const struct vrend_transfer_info *info,
                           UNUSED int transfer_mode)
{
   drm_log("unsupported");
   return -1;
}

static int
asahi_ccmd_nop(UNUSED struct drm_context *dctx, UNUSED const struct drm_ccmd_req *hdr)
{
   return 0;
}

static int
asahi_ccmd_ioctl_simple(struct drm_context *dctx, const struct drm_ccmd_req *hdr)
{
   const struct asahi_ccmd_ioctl_simple_req *req = to_asahi_ccmd_ioctl_simple_req(hdr);
   unsigned payload_len = _IOC_SIZE(req->cmd);
   unsigned req_len = size_add(sizeof(*req), payload_len);

   if (hdr->len != req_len) {
      drm_log("%u != %u", hdr->len, req_len);
      return -EINVAL;
   }

   /* Apply a reasonable upper bound on ioctl size: */
   if (payload_len > 128) {
      drm_log("invalid ioctl payload length: %u", payload_len);
      return -EINVAL;
   }

   /* Allow-list of supported ioctls: */
   unsigned iocnr = _IOC_NR(req->cmd) - DRM_COMMAND_BASE;
   switch (iocnr) {
   case DRM_ASAHI_GET_PARAMS:
   case DRM_ASAHI_VM_CREATE:
   case DRM_ASAHI_VM_DESTROY:
   case DRM_ASAHI_QUEUE_CREATE:
   case DRM_ASAHI_QUEUE_DESTROY:
      break;
   default:
      drm_log("invalid ioctl: %08x (%u)", req->cmd, iocnr);
      return -EINVAL;
   }

   struct asahi_ccmd_ioctl_simple_rsp *rsp;
   unsigned rsp_len = sizeof(*rsp);

   if (req->cmd & IOC_OUT)
      rsp_len = size_add(rsp_len, payload_len);

   rsp = drm_context_rsp(dctx, hdr, rsp_len);

   if (!rsp)
      return -ENOMEM;

   /* Copy the payload because the kernel can write (if IOC_OUT bit
    * is set) and to avoid casting away the const:
    */
   char payload[payload_len];
   memcpy(payload, req->payload, payload_len);

   rsp->ret = drmIoctl(dctx->fd, req->cmd, payload);

   if (req->cmd & IOC_OUT)
      memcpy(rsp->payload, payload, payload_len);

   return 0;
}

static int
asahi_ccmd_get_params(struct drm_context *dctx, const struct drm_ccmd_req *hdr)
{
   struct asahi_ccmd_get_params_req *req = to_asahi_ccmd_get_params_req(hdr);
   unsigned req_len = sizeof(*req);

   if (hdr->len != req_len) {
      drm_log("%u != %u", hdr->len, req_len);
      return -EINVAL;
   }

   struct asahi_ccmd_get_params_rsp *rsp;
   unsigned rsp_len = sizeof(*rsp);

   rsp = drm_context_rsp(dctx, hdr, rsp_len);

   if (!rsp)
      return -ENOMEM;

   req->params.pointer = (uint64_t)&rsp->params;

   rsp->ret = drmIoctl(dctx->fd, DRM_IOCTL_ASAHI_GET_PARAMS, &req->params);

   return 0;
}

static int
asahi_ccmd_gem_new(struct drm_context *dctx, const struct drm_ccmd_req *hdr)
{
   const struct asahi_ccmd_gem_new_req *req = to_asahi_ccmd_gem_new_req(hdr);
   int ret = 0;

   if (!drm_valid_blob_id(dctx, req->blob_id)) {
      drm_log("Invalid blob_id %u\n", req->blob_id);
      ret = -EINVAL;
      goto out_error;
   }

   /*
    * First part, allocate the GEM bo:
    */
   struct drm_asahi_gem_create gem_create = {
      .flags = req->flags,
      .vm_id = 0,
      .size = req->size,
   };

   ret = drmIoctl(dctx->fd, DRM_IOCTL_ASAHI_GEM_CREATE, &gem_create);
   if (ret) {
      drm_log("GEM_CREATE failed: %d (%s)\n", ret, strerror(errno));
      goto out_error;
   }

   /*
    * Second part, bind:
    */

   struct drm_asahi_gem_bind gem_bind = {
      .op = ASAHI_BIND_OP_BIND,
      .flags = req->bind_flags,
      .handle = gem_create.handle,
      .vm_id = req->vm_id,
      .offset = 0,
      .range = req->size,
      .addr = req->addr,
   };

   ret = drmIoctl(dctx->fd, DRM_IOCTL_ASAHI_GEM_BIND, &gem_bind);
   if (ret) {
      drm_log("DRM_IOCTL_ASAHI_GEM_BIND failed: (handle=%d)\n", gem_create.handle);
      goto out_close;
   }

   int prime_fd = -1;
   ret =
      drmPrimeHandleToFD(dctx->fd, gem_create.handle, DRM_CLOEXEC | DRM_RDWR, &prime_fd);
   if (ret) {
      fprintf(stderr, "Couldn't obtain PRIME_FD for handle: %d\n", gem_create.handle);
      return ret;
   }

   /*
    * And then finally create our drm_object for tracking the resource,
    * and add to blob table:
    */
   struct drm_object *obj =
      drm_object_create(gem_create.handle, req->flags, req->size, prime_fd);

   if (!obj) {
      ret = -ENOMEM;
      goto out_close;
   }

   drm_object_set_blob_id(dctx, obj, req->blob_id);

   drm_dbg("obj=%p, blob_id=%u, handle=%u\n", obj, obj->blob_id, obj->handle);

   return 0;

out_close:
   gem_close(dctx->fd, gem_create.handle);
out_error:
   if (dctx->shmem)
      dctx->shmem->async_error++;
   return ret;
}

static int
asahi_ccmd_gem_bind(struct drm_context *dctx, const struct drm_ccmd_req *hdr)
{
   const struct asahi_ccmd_gem_bind_req *req = to_asahi_ccmd_gem_bind_req(hdr);
   struct drm_object *obj = drm_get_object_from_res_id(dctx, req->res_id);
   int ret = 0;

   if (!obj) {
      drm_log("Could not lookup obj: res_id=%u", req->res_id);
      return -ENOENT;
   }

   drm_dbg("gem_bind: handle=%d\n", obj->handle);

   struct drm_asahi_gem_bind gem_bind = {
      .op = req->op,
      .flags = req->flags,
      .handle = obj->handle,
      .vm_id = req->vm_id,
      .offset = 0,
      .range = req->size,
      .addr = req->addr,
   };

   ret = drmIoctl(dctx->fd, DRM_IOCTL_ASAHI_GEM_BIND, &gem_bind);
   if (ret) {
      drm_log("DRM_IOCTL_ASAHI_GEM_BIND failed: (handle=%d)\n", obj->handle);
   }

   return ret;
}

static int
asahi_ccmd_submit(struct drm_context *dctx, const struct drm_ccmd_req *hdr)
{
   const struct asahi_ccmd_submit_req *req = to_asahi_ccmd_submit_req(hdr);
   int ret = 0;

   if (req->cmd_type != DRM_ASAHI_CMD_RENDER) {
      drm_log("!DRM_ASAHI_CMD_RENDER\n");
      return -EINVAL;
   }

   uint32_t syncobj;
   ret = drmSyncobjCreate(dctx->fd, 0, &syncobj);
   if (ret) {
      fprintf(stderr, "Can't create syncobj: %d", ret);
      return -EINVAL;
   }
   struct drm_asahi_sync out_sync = {
      .sync_type = DRM_ASAHI_SYNC_SYNCOBJ,
      .handle = syncobj,
   };

   struct drm_asahi_sync *in_syncs = (struct drm_asahi_sync *)req->payload;
   uint64_t cmd_buffer =
      (uint64_t)&req->payload + (sizeof(struct drm_asahi_sync) * req->in_sync_count);

   int obj_handles[16];
   for (int i = 0; i < req->in_sync_count; i++) {
      struct drm_object *obj = drm_get_object_from_res_id(dctx, in_syncs[i].handle);
      if (!obj) {
         fprintf(stderr, "couldn't find object for handle=%d\n", in_syncs[i].handle);
         return -EINVAL;
      }

      struct dma_buf_export_sync_file export_sync_file_ioctl = {
         .flags = DMA_BUF_SYNC_RW,
         .fd = -1,
      };

      ret =
         drmIoctl(obj->prime_fd, DMA_BUF_IOCTL_EXPORT_SYNC_FILE, &export_sync_file_ioctl);
      if (ret < 0) {
         fprintf(stderr, "couldn't export sync file form handle=%d: %d\n", obj->handle,
                 errno);
         continue;
         return -EINVAL;
      }

      obj_handles[i] = obj->prime_fd;
      int in_sync_fd = export_sync_file_ioctl.fd;

      /* Create a new syncobj */
      uint32_t sync_handle;
      ret = drmSyncobjCreate(dctx->fd, 0, &sync_handle);
      if (ret < 0) {
         fprintf(stderr, "couldn't create syncobj for handle\n");
         return -EINVAL;
      }

      /* Import the sync file into it */
      ret = drmSyncobjImportSyncFile(dctx->fd, sync_handle, in_sync_fd);
      if (ret < 0) {
         fprintf(stderr, "couldn't import syncobj for handle\n");
         return -EINVAL;
      }

      close(in_sync_fd);

      in_syncs[i].handle = sync_handle;
   }

   struct drm_asahi_cmd_render *crender = (struct drm_asahi_cmd_render *)cmd_buffer;
   crender->fragment_attachments = cmd_buffer + req->cmdbuf_size;

   drm_dbg("%s: fragment_count: %d", __func__, crender->fragment_attachment_count);

   struct drm_asahi_command cmd = {
      .cmd_type = req->cmd_type,
      .flags = req->flags,
      .cmd_buffer = (uint64_t)(uintptr_t)cmd_buffer,
      .cmd_buffer_size = req->cmdbuf_size,
      .result_offset = req->result_offset,
      .result_size = req->result_size,
      .barriers = { DRM_ASAHI_BARRIER_NONE, DRM_ASAHI_BARRIER_NONE },
   };

   for (int i = 0; i < DRM_ASAHI_SUBQUEUE_COUNT; i++) {
      cmd.barriers[i] = req->barriers[i];
   }

   struct drm_asahi_submit submit = {
      .flags = 0,
      .queue_id = req->queue_id,
      .result_handle = req->result_handle,
      .in_sync_count = req->in_sync_count,
      .out_sync_count = 1,
      .command_count = 1,
      .in_syncs = (uint64_t)(uintptr_t)in_syncs,
      .out_syncs = (uint64_t)(uintptr_t)&out_sync,
      .commands = (uint64_t)(uintptr_t)&cmd,
   };

   ret = drmIoctl(dctx->fd, DRM_IOCTL_ASAHI_SUBMIT, &submit);
   if (ret) {
      fprintf(stderr, "DRM_IOCTL_ASAHI_SUBMIT failed: %d\n", ret);
      return ret;
   }

   int fd = -1;
   ret = drmSyncobjExportSyncFile(dctx->fd, syncobj, &fd);
   if (ret) {
      fprintf(stderr, "Can't export file: %d", ret);
      return ret;
   }

   for (unsigned i = 0; i < req->in_sync_count; i++) {
      /* Free the in_sync handle we just acquired */
      ret = drmSyncobjDestroy(dctx->fd, in_syncs[i].handle);
      if (ret < 0) {
         fprintf(stderr, "couldn't destroy syncobj for handle=%d\n", in_syncs[i].handle);
         continue;
      }

      struct dma_buf_import_sync_file import_sync_file_ioctl = {
         .flags = DMA_BUF_SYNC_WRITE,
         .fd = fd,
      };

      ret = drmIoctl(obj_handles[i], DMA_BUF_IOCTL_IMPORT_SYNC_FILE,
                     &import_sync_file_ioctl);
      if (ret < 0) {
         fprintf(stderr, "couldn't import syncfile into syncobj for handle=%d\n",
                 in_syncs[i].handle);
         continue;
      }
   }

   drm_timeline_set_last_fence_fd(&dctx->timelines[0], fd);

   return ret;
}

static int
asahi_ccmd_syncobj_wait(struct drm_context *dctx, const struct drm_ccmd_req *hdr)
{
   const struct asahi_ccmd_syncobj_wait_req *req = to_asahi_ccmd_syncobj_wait_req(hdr);
   int ret = 0;

   struct drm_syncobj_wait wait = {
      .handles = (intptr_t)req->payload,
      .flags = req->flags,
      .timeout_nsec = req->timeout_nsecs,
      .count_handles = req->handle_count,
   };

   struct asahi_ccmd_syncobj_wait_rsp *rsp;
   unsigned rsp_len = sizeof(*rsp);

   rsp = drm_context_rsp(dctx, hdr, rsp_len);

   if (!rsp)
      return -ENOMEM;

   ret = drmIoctl(dctx->fd, DRM_IOCTL_SYNCOBJ_WAIT, &wait);
   rsp->ret = ret;
   if (ret < 0) {
      rsp->ret = errno;
      fprintf(stderr, "DRM_IOCTL_SYNCOBJ_WAIT failed: %d, %m\n", ret);
      return 0;
   }

   rsp->first_signaled = wait.first_signaled;

   return 0;
}

static int
asahi_ccmd_syncobj_create(struct drm_context *dctx, const struct drm_ccmd_req *hdr)
{
   const struct asahi_ccmd_syncobj_create_req *req =
      to_asahi_ccmd_syncobj_create_req(hdr);
   int ret = 0;

   struct drm_syncobj_create create = {
      .handle = 0,
      .flags = req->flags,
   };

   struct asahi_ccmd_syncobj_create_rsp *rsp;
   unsigned rsp_len = sizeof(*rsp);

   rsp = drm_context_rsp(dctx, hdr, rsp_len);

   if (!rsp)
      return -ENOMEM;

   ret = drmIoctl(dctx->fd, DRM_IOCTL_SYNCOBJ_CREATE, &create);
   rsp->ret = ret;
   if (ret < 0) {
      drm_log("DRM_IOCTL_SYNCOBJ_CREATE failed: %d\n", ret);
      return ret;
   }

   rsp->handle = create.handle;

   drm_dbg("%s: new handle=%d\n", __func__, rsp->handle);

   return 0;
}

static const struct ccmd {
   const char *name;
   int (*handler)(struct drm_context *dctx, const struct drm_ccmd_req *hdr);
   size_t size;
} ccmd_dispatch[] = {
#define HANDLER(N, n)                                                                    \
   [ASAHI_CCMD_##N] = { #N, asahi_ccmd_##n, sizeof(struct asahi_ccmd_##n##_req) }
   HANDLER(NOP, nop),
   HANDLER(IOCTL_SIMPLE, ioctl_simple),
   HANDLER(GET_PARAMS, get_params),
   HANDLER(GEM_NEW, gem_new),
   HANDLER(GEM_BIND, gem_bind),
   HANDLER(SUBMIT, submit),
   HANDLER(SYNCOBJ_WAIT, syncobj_wait),
   HANDLER(SYNCOBJ_CREATE, syncobj_create),
};

static int
submit_cmd_dispatch(struct drm_context *dctx, const struct drm_ccmd_req *hdr)
{
   int ret;

   if (hdr->cmd >= ARRAY_SIZE(ccmd_dispatch)) {
      drm_log("invalid cmd: %u", hdr->cmd);
      return -EINVAL;
   }

   const struct ccmd *ccmd = &ccmd_dispatch[hdr->cmd];

   if (!ccmd->handler) {
      drm_log("no handler: %u", hdr->cmd);
      return -EINVAL;
   }

   drm_dbg("%s: hdr={cmd=%u, len=%u, seqno=%u, rsp_off=0x%x)", ccmd->name, hdr->cmd,
           hdr->len, hdr->seqno, hdr->rsp_off);

   /* If the request length from the guest is smaller than the expected
    * size, ie. newer host and older guest, we need to make a copy of
    * the request with the new fields at the end zero initialized.
    */
   if (ccmd->size > hdr->len) {
      uint8_t buf[ccmd->size];

      memcpy(&buf[0], hdr, hdr->len);
      memset(&buf[hdr->len], 0, ccmd->size - hdr->len);

      ret = ccmd->handler(dctx, (struct drm_ccmd_req *)buf);
   } else {
      ret = ccmd->handler(dctx, hdr);
   }

   if (ret) {
      drm_log("%s: dispatch failed: %d (%s)", ccmd->name, ret, strerror(errno));
      return ret;
   }

   /* If the response length from the guest is smaller than the
    * expected size, ie. newer host and older guest, then a shadow
    * copy is used, and we need to copy back to the actual rsp
    * buffer.
    */
   struct drm_ccmd_rsp *rsp = drm_context_rsp_noshadow(dctx, hdr);
   if (dctx->current_rsp && (dctx->current_rsp != rsp)) {
      unsigned len = rsp->len;
      memcpy(rsp, dctx->current_rsp, len);
      rsp->len = len;
      free(dctx->current_rsp);
   }
   dctx->current_rsp = NULL;

   /* Note that commands with no response, like SET_DEBUGINFO, could
    * be sent before the shmem buffer is allocated:
    */
   if (dctx->shmem) {
      /* TODO better way to do this?  We need ACQ_REL semanatics (AFAIU)
       * to ensure that writes to response buffer are visible to the
       * guest process before the update of the seqno.  Otherwise we
       * could just use p_atomic_set.
       */
      uint32_t seqno = hdr->seqno;
      drm_log("updating seqno=%d\n", seqno);
      p_atomic_xchg(&dctx->shmem->seqno, seqno);
   }

   return 0;
}

static int
asahi_renderer_submit_cmd(struct virgl_context *vctx, const void *_buffer, size_t size)
{
   struct drm_context *dctx = to_drm_context(vctx);
   const uint8_t *buffer = _buffer;

   while (size >= sizeof(struct drm_ccmd_req)) {
      const struct drm_ccmd_req *hdr = (const struct drm_ccmd_req *)buffer;

      /* Sanity check first: */
      if ((hdr->len > size) || (hdr->len < sizeof(*hdr)) || (hdr->len % 4)) {
         fprintf(stderr, "bad size, %u vs %zu (%u)", hdr->len, size, hdr->cmd);
         goto cont;
      }

      if (hdr->rsp_off % 4) {
         fprintf(stderr, "bad rsp_off, %u", hdr->rsp_off);
         goto cont;
      }

      int ret = submit_cmd_dispatch(dctx, hdr);
      if (ret) {
         fprintf(stderr, "dispatch failed: %d (%u)", ret, hdr->cmd);
      }

   cont:
      buffer += hdr->len;
      size -= hdr->len;
   }

   if (size > 0) {
      fprintf(stderr, "bad size, %zu trailing bytes", size);
      return -EINVAL;
   }

   return 0;
}

static int
asahi_renderer_get_fencing_fd(struct virgl_context *vctx)
{
   struct drm_context *dctx = to_drm_context(vctx);
   return dctx->eventfd;
}

static void
asahi_renderer_retire_fences(UNUSED struct virgl_context *vctx)
{
   /* No-op as VIRGL_RENDERER_ASYNC_FENCE_CB is required */
}

static void
asahi_renderer_fence_retire(struct virgl_context *vctx, uint32_t ring_idx,
                            uint64_t fence_id)
{
   // TODO - destroy syncobj?
   vctx->fence_retire(vctx, ring_idx, fence_id);
}

static int
asahi_renderer_submit_fence(struct virgl_context *vctx, uint32_t flags, uint32_t ring_idx,
                            uint64_t fence_id)
{
   struct drm_context *dctx = to_drm_context(vctx);

   drm_dbg(stderr, "flags=0x%x, ring_idx=%" PRIu32 ", fence_id=%" PRIu64, flags, ring_idx,
           fence_id);

   /* timeline is ring_idx-1 (because ring_idx 0 is host CPU timeline) */
   if (ring_idx > nr_timelines) {
      drm_log("invalid ring_idx: %" PRIu32, ring_idx);
      return -EINVAL;
   }

   /* ring_idx zero is used for the guest to synchronize with host CPU,
    * meaning by the time ->submit_fence() is called, the fence has
    * already passed.. so just immediate signal:
    */
   if (ring_idx == 0) {
      vctx->fence_retire(vctx, ring_idx, fence_id);
      return 0;
   }

   return drm_timeline_submit_fence(&dctx->timelines[0], flags, fence_id);
}

struct virgl_context *
asahi_renderer_create(int fd)
{
   struct drm_context *dctx;

   drm_log("");

   dctx = calloc(1, sizeof(*dctx) + (nr_timelines * sizeof(dctx->timelines[0])));
   if (!dctx)
      return NULL;

   dctx->fd = fd;

   /* Indexed by blob_id, but only lower 32b of blob_id are used: */
   dctx->blob_table = _mesa_hash_table_create_u32_keys(NULL);
   /* Indexed by res_id: */
   dctx->resource_table = _mesa_hash_table_create_u32_keys(NULL);
   /* Indexed by submitqueue-id: */
   dctx->sq_to_ring_idx_table = _mesa_hash_table_create_u32_keys(NULL);

   dctx->eventfd = create_eventfd(0);

   drm_timeline_init(&dctx->timelines[0], &dctx->base, "asahi-sync", dctx->eventfd, 1,
                     asahi_renderer_fence_retire);

   dctx->base.destroy = asahi_renderer_destroy;
   dctx->base.attach_resource = drm_renderer_attach_resource;
   dctx->base.detach_resource = asahi_renderer_detach_resource;
   dctx->base.export_opaque_handle = drm_renderer_export_opaque_handle;
   dctx->base.transfer_3d = asahi_renderer_transfer_3d;
   dctx->base.get_blob = drm_renderer_get_blob;
   dctx->base.submit_cmd = asahi_renderer_submit_cmd;
   dctx->base.get_fencing_fd = asahi_renderer_get_fencing_fd;
   dctx->base.retire_fences = asahi_renderer_retire_fences;
   dctx->base.submit_fence = asahi_renderer_submit_fence;
   dctx->base.supports_fence_sharing = true;

   return &dctx->base;
}
