/*
 * Copyright 2023 Red Hat, Inc.
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: MIT
 */

#ifndef DRM_CONTEXT_H_
#define DRM_CONTEXT_H_

#include "virgl_context.h"

#include "drm_fence.h"

#define DEFINE_CAST(parent, child)                                                       \
   static inline struct child *to_##child(const struct parent *x)                        \
   {                                                                                     \
      return (struct child *)x;                                                          \
   }

/**
 * Defines the layout of shmem buffer used for host->guest communication.
 */
struct drm_shmem {
   /**
    * The sequence # of last cmd processed by the host
    */
   uint32_t seqno;

   /**
    * Offset to the start of rsp memory region in the shmem buffer.  This
    * is set by the host when the shmem buffer is allocated, to allow for
    * extending the shmem buffer with new fields.  The size of the rsp
    * memory region is the size of the shmem buffer (controlled by the
    * guest) minus rsp_mem_offset.
    *
    * The guest should use the drm_shmem_has_field() macro to determine
    * if the host supports a given field, ie. to handle compatibility of
    * newer guest vs older host.
    *
    * Making the guest userspace responsible for backwards compatibility
    * simplifies the host VMM.
    */
   uint32_t rsp_mem_offset;

#define drm_shmem_has_field(shmem, field)                                                \
   ({                                                                                    \
      struct drm_shmem *_shmem = (shmem);                                                \
      (_shmem->rsp_mem_offset > offsetof(struct drm_shmem, field));                      \
   })

   /**
    * Counter that is incremented on asynchronous errors, like SUBMIT
    * or GEM_NEW failures.  The guest should treat errors as context-
    * lost.
    */
   uint32_t async_error;

   /**
    * Counter that is incremented on global fault (see DRM_PARAM_FAULTS)
    */
   uint32_t global_faults;
};

struct drm_ccmd_req {
   uint32_t cmd;
   uint32_t len;
   uint32_t seqno;

   /* Offset into shmem ctrl buffer to write response.  The host ensures
    * that it doesn't write outside the bounds of the ctrl buffer, but
    * otherwise it is up to the guest to manage allocation of where responses
    * should be written in the ctrl buf.
    */
   uint32_t rsp_off;
};

struct drm_ccmd_rsp {
   uint32_t len;
};

/**
 * A single context (from the PoV of the virtio-gpu protocol) maps to
 * a single drm device open.  Other drm constructs (ie. submitqueue) are
 * opaque to the protocol.
 *
 * Typically each guest process will open a single virtio-gpu "context".
 * The single drm device open maps to an individual GEM address_space on
 * the kernel side, providing GPU address space isolation between guest
 * processes.
 *
 * GEM buffer objects are tracked via one of two id's:
 *  - resource-id:  global, assigned by guest kernel
 *  - blob-id:      context specific, assigned by guest userspace
 *
 * The blob-id is used to link the bo created via the corresponding ioctl
 * and the get_blob() cb. It is unused in the case of a bo that is
 * imported from another context.  An object is added to the blob table
 * in GEM_NEW and removed in ctx->get_blob() (where it is added to
 * resource_table). By avoiding having an obj in both tables, we can
 * safely free remaining entries in either hashtable at context teardown.
 */
struct drm_context {
   struct virgl_context base;

   struct drm_shmem *shmem;
   uint8_t *rsp_mem;
   uint32_t rsp_mem_sz;

   struct drm_ccmd_rsp *current_rsp;

   int fd;

   struct hash_table *blob_table;
   struct hash_table *resource_table;

   /**
    * Maps submit-queue id to ring_idx
    */
   struct hash_table *sq_to_ring_idx_table;

   int eventfd;

   /**
    * Indexed by ring_idx-1, which is the same as the submitqueue priority+1.
    * On the kernel side, there is some drm_sched_entity per {drm_file, prio}
    * tuple, and the sched entity determines the fence timeline, ie. submits
    * against a single sched entity complete in fifo order.
    */
   struct drm_timeline timelines[];
};
DEFINE_CAST(virgl_context, drm_context)

void drm_renderer_unmap_blob(struct drm_context *dctx);

void drm_renderer_attach_resource(struct virgl_context *vctx, struct virgl_resource *res);

void drm_renderer_detach_resource(struct virgl_context *vctx, struct virgl_resource *res,
                                  int (*gem_close)(int, uint32_t));

enum virgl_resource_fd_type drm_renderer_export_opaque_handle(struct virgl_context *vctx,
                                                              struct virgl_resource *res,
                                                              int *out_fd);

int drm_renderer_get_blob(struct virgl_context *vctx, uint32_t res_id, uint64_t blob_id,
                          uint64_t blob_size, uint32_t blob_flags,
                          struct virgl_context_blob *blob);

void *drm_context_rsp_noshadow(struct drm_context *dctx, const struct drm_ccmd_req *hdr);

void *drm_context_rsp(struct drm_context *dctx, const struct drm_ccmd_req *hdr,
                      unsigned len);

#endif // DRM_CONTEXT_H_
