#include "drm_object.h"
#include "drm_util.h"

struct drm_object *
drm_object_create(uint32_t handle, uint32_t flags, uint32_t size, uint32_t prime_fd)
{
   struct drm_object *obj = calloc(1, sizeof(*obj));

   if (!obj)
      return NULL;

   obj->handle = handle;
   obj->flags = flags;
   obj->size = size;
   obj->prime_fd = prime_fd;

   return obj;
}

bool
drm_valid_blob_id(struct drm_context *dctx, uint32_t blob_id)
{
   /* must be non-zero: */
   if (blob_id == 0)
      return false;

   /* must not already be in-use: */
   if (drm_table_search(dctx->blob_table, blob_id))
      return false;

   return true;
}

void
drm_object_set_blob_id(struct drm_context *dctx, struct drm_object *obj, uint32_t blob_id)
{
   assert(drm_valid_blob_id(dctx, blob_id));

   obj->blob_id = blob_id;
   _mesa_hash_table_insert(dctx->blob_table, (void *)(uintptr_t)obj->blob_id, obj);
}

bool
drm_valid_res_id(struct drm_context *dctx, uint32_t res_id)
{
   return !drm_table_search(dctx->resource_table, res_id);
}

void
drm_object_set_res_id(struct drm_context *dctx, struct drm_object *obj, uint32_t res_id)
{
   assert(valid_res_id(dctx, res_id));

   obj->res_id = res_id;
   _mesa_hash_table_insert(dctx->resource_table, (void *)(uintptr_t)obj->res_id, obj);
}

void
drm_remove_object(struct drm_context *dctx, struct drm_object *obj)
{
   drm_dbg("obj=%p, blob_id=%u, res_id=%u", obj, obj->blob_id, obj->res_id);
   _mesa_hash_table_remove_key(dctx->resource_table, (void *)(uintptr_t)obj->res_id);
}

struct drm_object *
drm_retrieve_object_from_blob_id(struct drm_context *dctx, uint64_t blob_id)
{
   assert((blob_id >> 32) == 0);
   uint32_t id = blob_id;
   struct hash_entry *entry = drm_table_search(dctx->blob_table, id);
   if (!entry)
      return NULL;
   struct drm_object *obj = entry->data;
   _mesa_hash_table_remove(dctx->blob_table, entry);
   return obj;
}

struct drm_object *
drm_get_object_from_res_id(struct drm_context *dctx, uint32_t res_id)
{
   const struct hash_entry *entry = drm_table_search(dctx->resource_table, res_id);
   return likely(entry) ? entry->data : NULL;
}

uint32_t
drm_handle_from_res_id(struct drm_context *dctx, uint32_t res_id)
{
   struct drm_object *obj = drm_get_object_from_res_id(dctx, res_id);
   if (!obj)
      return 0; /* zero is an invalid GEM handle */
   return obj->handle;
}
